const renderEngine = require('../render-engine');
const gfx = renderEngine.gfx;

const isIOS14Device = cc.sys.os === cc.sys.OS_IOS && cc.sys.isBrowser && cc.sys.isMobile && /iPhone OS 14/.test(window.navigator.userAgent);


let MeshBuffer = cc.Class({
    name: 'cc.MeshBuffer',
    ctor (batcher, vertexFormat) {
        this.byteStart = 0;
        this.byteOffset = 0;
        this.indiceStart = 0;
        this.indiceOffset = 0;
        this.vertexStart = 0;
        this.vertexOffset = 0;

        this._vertexFormat = vertexFormat;
        this._vertexBytes = this._vertexFormat._bytes;

        this._arrOffset = 0;
        this._vbArr = [];
        this._vb = new gfx.VertexBuffer(
            batcher._device,
            vertexFormat,
            gfx.USAGE_DYNAMIC,
            new ArrayBuffer(),
            0
        );
        this._vbArr[0] = this._vb;

        this._ibArr = [];
        this._ib = new gfx.IndexBuffer(
            batcher._device,
            gfx.INDEX_FMT_UINT16,
            gfx.USAGE_STATIC,
            new ArrayBuffer(),
            0
        );
        this._ibArr[0] = this._ib;

        this._vData = null;
        this._uintVData = null;
        this._iData = null;

        this._batcher = batcher;

        this._initVDataCount = 256 * vertexFormat._bytes;// actually 256 * 4 * (vertexFormat._bytes / 4)
        this._initIDataCount = 256 * 6;
        
        this._offsetInfo = {
            byteOffset : 0,
            vertexOffset : 0,
            indiceOffset : 0
        }
        this._reallocBuffer();
    },

    uploadData () {
        if (this.byteOffset === 0 || !this._dirty) {
            return;
        }

        // update vertext data
        let vertexsData = new Float32Array(this._vData.buffer, 0, this.byteOffset >> 2);
        let indicesData = new Uint16Array(this._iData.buffer, 0, this.indiceOffset);

        let vb = this._vb;
        vb.update(0, vertexsData);

        let ib = this._ib;
        ib.update(0, indicesData);

        this._dirty = false;
    },

    switchBuffer () {
        let offset = ++this._arrOffset;

        this.byteStart = 0;
        this.byteOffset = 0;
        this.vertexStart = 0;
        this.vertexOffset = 0;
        this.indiceStart = 0;
        this.indiceOffset = 0;

        if (offset < this._vbArr.length) {
            this._vb = this._vbArr[offset];
            this._ib = this._ibArr[offset];
        } else {

            this._vb = new gfx.VertexBuffer(
                this._batcher._device,
                this._vertexFormat,
                gfx.USAGE_DYNAMIC,
                new ArrayBuffer(),
                0
            );
            this._vbArr[offset] = this._vb;
            this._vb._bytes = this._vData.byteLength;

            this._ib = new gfx.IndexBuffer(
                this._batcher._device,
                gfx.INDEX_FMT_UINT16,
                gfx.USAGE_STATIC,
                new ArrayBuffer(),
                0
            );
            this._ibArr[offset] = this._ib;
            this._ib._bytes = this._iData.byteLength;
        }
    },

    checkAndSwitchBuffer (vertexCount) {

        if (isIOS14Device) {
            if (this.vertexOffset + vertexCount > 65535) {
                this.uploadData();
                this._batcher._flush();
            }
        }else{
            if (this.vertexOffset + vertexCount > 65535) {
                this.uploadData();
                this._batcher._flush();
                this.switchBuffer();
            }
        }
    },

    requestStatic (vertexCount, indiceCount) {

        this.checkAndSwitchBuffer(vertexCount);
        //字节偏移 = 当前字节偏移+顶点数*顶点字节数
        let byteOffset = this.byteOffset + vertexCount * this._vertexBytes;
        //索引偏移 = 当前索引偏移 + 索引数
        let indiceOffset = this.indiceOffset + indiceCount;
        //数据的长度
        let byteLength = this._vData.byteLength;
        //索引数据的长度
        let indiceLength = this._iData.length;
        //字节偏移>顶点数据的长度   或者   索引偏移>索引数据的长度
        if (byteOffset > byteLength || indiceOffset > indiceLength) {
            //循环处理：顶点数据的长度<字节偏移 || 索引数据长度<索引偏移
            while (byteLength < byteOffset || indiceLength < indiceOffset) {
                //顶点数据个数
                this._initVDataCount *= 2;
                //索引数据个数
                this._initIDataCount *= 2;
                //顶点数据字节数 = 个数*字节数
                byteLength = this._initVDataCount * 4;
                //索引数 = 索引个数
                indiceLength = this._initIDataCount;
            }
            //顶点数大于当有buffer长度时
            this._reallocBuffer();
        }

        let offsetInfo = this._offsetInfo;
        offsetInfo.vertexOffset = this.vertexOffset;
        this.vertexOffset += vertexCount;

        offsetInfo.indiceOffset = this.indiceOffset;
        this.indiceOffset += indiceCount;

        offsetInfo.byteOffset = this.byteOffset;
        this.byteOffset = byteOffset;

        this._dirty = true;
    },

    request (vertexCount, indiceCount) {
        if (this._batcher._buffer !== this) {
            this._batcher._flush();
            this._batcher._buffer = this;
        }

        this.requestStatic(vertexCount, indiceCount);
        return this._offsetInfo;
    },
    
    _reallocBuffer () {
        this._reallocVData(true);
        this._reallocIData(true);
    },

    _reallocVData (copyOldData) {
        let oldVData;
        if (this._vData) {
            //当前的数据
            oldVData = new Uint8Array(this._vData.buffer);
        }

        //32位的浮点型数组（顶点数）
        this._vData = new Float32Array(this._initVDataCount);
        //32位的无符号整型数组（32位浮点型数组缓存）
        this._uintVData = new Uint32Array(this._vData.buffer);
        //8位的无符号整型数组（32位无符号整型数组缓存）
        let newData = new Uint8Array(this._uintVData.buffer);
        //如果有老数据并且需要复制数据
        if (oldVData && copyOldData) {
            //复制数据到新的数组
            for (let i = 0, l = oldVData.length; i < l; i++) {
                newData[i] = oldVData[i];
            }
        }
        //字节数
        this._vb._bytes = this._vData.byteLength;
    },

    _reallocIData (copyOldData) {
        let oldIData = this._iData;

        this._iData = new Uint16Array(this._initIDataCount);

        if (oldIData && copyOldData) {
            let iData = this._iData;
            for (let i = 0, l = oldIData.length; i < l; i++) {
                iData[i] = oldIData[i];
            }
        }

        this._ib._bytes = this._iData.byteLength;
    },

    reset () {
        this._arrOffset = 0;
        this._vb = this._vbArr[0];
        this._ib = this._ibArr[0];

        this.byteStart = 0;
        this.byteOffset = 0;
        this.indiceStart = 0;
        this.indiceOffset = 0;
        this.vertexStart = 0;
        this.vertexOffset = 0;

        this._dirty = false;
    },

    destroy () {
        for (let key in this._vbArr) {
            let vb = this._vbArr[key];
            vb.destroy();
        }
        this._vbArr = undefined;

        for (let key in this._ibArr) {
            let ib = this._ibArr[key];
            ib.destroy();
        }
        this._ibArr = undefined;

        this._ib = undefined;
        this._vb = undefined;
    }
});

cc.MeshBuffer = module.exports = MeshBuffer;
